# modules
from texttable import Texttable

class Subject:
    """
    This class stores the credit and grade point of a subject.
    """

    # constructor
    def __init__(self, credit: int, grade_point: int) -> None:

        # check if the credit and grade point are integers and are not negative
        if type(credit) != int:
            raise TypeError("Credit should be an integer.")
        if type(grade_point) != int:
            raise TypeError("Grade Point should be an integer.")
        if credit < 0 or grade_point < 0:
            raise ValueError("Credit and Grade Point cannot be negative.")
        
        
        self.credit = credit
        self.grade_point = grade_point

    def __str__(self) -> str:
        return "Credit: " + str(self.credit) + "\nGrade Point: " + str(self.grade_point)


class Gpa(Subject):
    """
    This class calculates the GPA of a student.
    It takes the credit and grade point of each subject as input and returns the GPA.
    """

    # Class variables
    # maintains no of subjects added
    no_of_subjects = 0

    # Constructor
    def __init__(self) -> None:
        self.gpa = 0.0
        self.numerator = 0.0
        self.denominator = 0.0
        self.sub=[]

    # Add a subject
    def addsubject(self, credit: int, grade_point: int) -> None:
        """
        addsubject(credit, grade_point)
        This function adds a subject to the list of subjects.
        """


        # limits on credit and grade point

        if credit > 8:
            raise ValueError("Credit cannot be greater than 8.")
        if grade_point > 10:
            raise ValueError("Grade Point cannot be greater than 10.")
            
        
        # Add subject to the list
        self.sub.append(Subject(credit, grade_point))
        self.no_of_subjects += 1

    # Calculate GPA
    def calc(self) -> float:
        """
        calc()
        This function calculates the GPA of the student.
        """

        if self.no_of_subjects == 0:
            raise ValueError("No subjects added yet. Run gpa.addsubject(credit, grade_point) first.")
            
        # Check if there is at least one subject with grade point greater than 5
        for i in range(self.no_of_subjects):
            greater_than_5 = False
            if self.sub[i].grade_point > 5:
                greater_than_5 = True
            else:
                pass
            
        if not greater_than_5:
            raise ValueError("No subjects with grade point greater than 5.")
        
        # Iterate through the list of subjects and calculate the numerator and denominator
        for i in range(self.no_of_subjects):

            # Check if the grade point is less than 5
            if self.sub[i].grade_point < 5:
                continue

            # Calculate the numerator and denominator
            self.numerator += self.sub[i].credit * self.sub[i].grade_point
            self.denominator += self.sub[i].credit
        
        # Calculate the GPA with the numerator and denominator
        self.gpa = self.numerator / self.denominator

        # Check if GPA is valid
        if self.gpa > 10 and self.gpa < 0:
            print("GPA is not valid. Please check your inputs.")
            return

        return self.gpa

    def display(self) -> None:
        """
        display()
        This function displays the number of subjects and the GPA.
        """


        # Check if at least one subject is added
        if self.no_of_subjects == 0:
            raise ValueError("No subjects added yet. Run gpa.addsubject(credit, grade_point) first.")
            

        # Check if GPA is calculated
        if self.gpa == 0.0:
            raise ValueError("No GPA calculated yet. Run gpa.calc() first.")
            
        
        # Display the number of subjects and the GPA
        print("No of subjects: ", self.no_of_subjects)
        print("GPA: ", self.gpa)

    def added(self) -> None:
        """
        added()
        This function displays the subjects added.
        """
        
        table = Texttable()

        table.header(["No.", "Credits", "Grade Points"])
        table.set_cols_dtype(['i', 'i', 'i'])
        table.set_cols_align(['r', 'r', 'r'])

        for i in range(self.no_of_subjects):
            table.add_row([i+1, self.sub[i].credit, self.sub[i].grade_point])

        table.set_deco(Texttable.HEADER)
        print('Subjects added:')
        print(table.draw())
        
    def removesubject(self, index: int) -> None:
        """
        removesubject(index)
        Index values: 1-n
        This function removes a subject from the list of subjects.
        """
        index = index-1
        # Check if the index is valid
        if index > self.no_of_subjects or index < 0:
            raise ValueError(f"Index out of range. Input Index: {index+1} Index Region: 0 to {self.no_of_subjects}")
            
        
        # Remove the subject from the list
        removed = self.sub.pop(index-1)
        self.no_of_subjects -= 1
        print(f'Subject {index + 1} {removed.credit, removed.grade_point} removed successfully.')

    def __repr__(self) -> str:
        return "GPA Calculator"
    
    def __str__(self) -> str:

        if self.no_of_subjects == 0:
            return "No subjects added yet. Run gpa.addsubject(credit, grade_point) first."
        
        return "No of subjects: " + str(self.no_of_subjects) + "GPA: " + str(self.gpa)
    

class Cgpa(Gpa):
    """
    This class calculates the CGPA of a student.
    It takes the GPA of each semester as input and returns the CGPA.
    """

    # Constructor
    def __init__(self):
        self.cgpa = 0.0
        self.total_numerator = 0.0
        self.total_denominator = 0
        self.semester = []
    

    # Add a semester gpa
    def addsemester(self, gpa: Gpa) -> None:
        """
        addsemester(gpa)
        This function adds a semester gpa to the list of semester gpa.
        """

        # Check if the input is of type Gpa else raise an error
        if not isinstance(gpa, Gpa):
            raise TypeError("Input must be of type Gpa.")

        # Check if the gpa is calculatable else raise an error
        if gpa.no_of_subjects <= 0:
            raise ValueError("No subjects added yet. Run gpa.addsubject(credit, grade_point) first.")
        


        gpa.calc()
        self.semester.append(gpa)
            
        
    # Calculate CGPA
    def calc(self) -> float:
        """
        calc()
        This function calculates the CGPA of the student.
        """

        if len(self.semester) == 0:
            raise ValueError("No semesters added yet. Run cgpa.addsemester(gpa) first.")
        
        # Iterate through the list of semester gpa and calculate the numerator and denominator
        for i in range(len(self.semester)):
            self.total_numerator += self.semester[i].numerator
            self.total_denominator += self.semester[i].denominator

        # Calculate the CGPA with the numerator and denominator
        self.cgpa = self.total_numerator / self.total_denominator

        # Check if CGPA is valid
        if self.cgpa > 10 and self.cgpa < 0:
            print("CGPA is not valid. Please check your inputs.")
            return

        return self.cgpa
    
    # Display CGPA
    def display(self) -> None:
        """
        display()
        This function displays the number of semesters and the CGPA.
        """

        if len(self.semester) == 0:
            raise ValueError("No semesters added yet. Run cgpa.addsemester(gpa) first.")
        if self.cgpa == 0.0:
            raise ValueError("No CGPA calculated yet. Run cgpa.calc() first.")
        
        print("No of semesters: ", len(self.semester))
        print("CGPA: ", self.cgpa)
    
    def added(self) -> None:
        """
        added()
        This function displays the semesters added.
        """
        
        table = Texttable()

        table.header(["No.", "GPA", "\u03A3 GP*C", "\u03A3 Credits"])
        table.set_cols_dtype(['i', 'f', 'i', 'i'])
        table.set_cols_align(['r', 'r', 'r', 'r'])

        for i in range(len(self.semester)):
            table.add_row([i+1, self.semester[i].gpa, self.semester[i].numerator, self.semester[i].denominator])

        table.set_deco(Texttable.HEADER)
        print('Semesters added:')
        print(table.draw())

    def removesemester(self, index: int) -> None:
        """
        removesemester(index)
        Index values: 1-n
        This function removes a semester from the list of semesters.
        """
        index = index-1
        # Check if the index is valid
        if index > len(self.semester) or index < 0:
            raise ValueError(f"Index out of range. Input Index: {index+1} Index Region: 0 to {len(self.semester)}")
            
        
        # Remove the semester from the list
        removed = self.semester.pop(index-1)
        print(f'Semester {index+1} {removed.gpa, removed.numerator, removed.denominator} removed successfully.')

    def __repr__(self) -> str:
        return "CGPA Calculator"

    def __str__(self) -> str:
        return "No of semesters: " + str(len(self.semester)) + "CGPA: " + str(self.cgpa)
        



import unittest

from src.rollno import isvalid, parse, get_dept_code, get_dept

class TestParse(unittest.TestCase):

    # ------------------------------------------------------

    # Valid rollno

    # ------------------------------------------------------


    def test_correct_rollno(self):
        self.assertTrue(isvalid('2021PECCB901'))
        self.assertTrue(isvalid('2021PECCB012'))
        self.assertTrue(isvalid('2021PECCB123'))
        self.assertTrue(isvalid('2021PECCB234'))
        self.assertTrue(isvalid('2021PECCB345'))
        self.assertTrue(isvalid('2021PECCB456'))
        self.assertTrue(isvalid('2021PECCB567'))
        self.assertTrue(isvalid('2021PECCB678'))
        self.assertTrue(isvalid('2021PECCB789'))
        self.assertTrue(isvalid('2021PECCB890'))

    def test_correct_dept(self):
        self.assertTrue(isvalid('2021PECCB000'))
        self.assertTrue(isvalid('2021PECCS000'))
        self.assertTrue(isvalid('2021PECIT000'))
        self.assertTrue(isvalid('2021PECEC000'))
        self.assertTrue(isvalid('2021PECEE000'))
        self.assertTrue(isvalid('2021PECCE000'))
        self.assertTrue(isvalid('2021PECME000'))
        self.assertTrue(isvalid('2021PECAD000'))
        self.assertTrue(isvalid('2021PECCC000'))

    def test_correct_year(self):
        self.assertTrue(isvalid('2018PECCB000'))
        self.assertTrue(isvalid('2019PECCB000'))
        self.assertTrue(isvalid('2020PECCB000'))
        self.assertTrue(isvalid('2021PECCB000'))
        self.assertTrue(isvalid('2022PECCB000'))
        self.assertTrue(isvalid('2023PECCB000'))
        self.assertTrue(isvalid('2024PECCB000'))
        self.assertTrue(isvalid('2025PECCB000'))
        self.assertTrue(isvalid('2026PECCB000'))
        self.assertTrue(isvalid('2027PECCB000'))
        self.assertTrue(isvalid('2028PECCB000'))
        self.assertTrue(isvalid('2029PECCB000'))
        self.assertTrue(isvalid('2030PECCB000'))

    def test_lettercase(self):
        self.assertTrue(isvalid('2021peccb000'))
        self.assertTrue(isvalid('2021PECCB000'))
        self.assertTrue(isvalid('2021PeCcB000'))
        self.assertTrue(isvalid('2021pEcCb000'))

    def test_inisvalid_rollno(self):
        self.assertFalse(isvalid('2021PECCB00'))
        self.assertFalse(isvalid('2021PECCB0000'))
        self.assertFalse(isvalid('2021PECCB00000'))
        self.assertFalse(isvalid('2021PECCB000000'))
        self.assertFalse(isvalid('2021PECCB0000000'))
        self.assertFalse(isvalid('2021PECCBXXX'))
        self.assertFalse(isvalid('2021PECCB000X'))
        self.assertFalse(isvalid('2021PECB00X0'))
        self.assertFalse(isvalid('2021PECB00X0'))
    
    def test_inisvalid_dept(self):
        self.assertFalse(isvalid('2021PEC11000'))
        self.assertFalse(isvalid('2021PEC22000'))
        self.assertFalse(isvalid('2021PEC33000'))
        self.assertFalse(isvalid('2021PEC44000'))
        self.assertFalse(isvalid('2021PEC55000'))
        self.assertFalse(isvalid('2021PEC66000'))
        self.assertFalse(isvalid('2021PEC000'))
        self.assertFalse(isvalid('2021PEC0B000'))
        self.assertFalse(isvalid('2021PEC1B000'))
        self.assertFalse(isvalid('2021PEC2B000'))
        self.assertFalse(isvalid('2021PEC3B000'))
        self.assertFalse(isvalid('2021PEC4B000'))
        self.assertFalse(isvalid('2021PEC5B000'))
        self.assertFalse(isvalid('2021PEC6B000'))
        self.assertFalse(isvalid('2021PEC7B000'))
        self.assertFalse(isvalid('2021PEC8B000'))
        self.assertFalse(isvalid('2021PEC9B000'))

    def test_inisvalid_year(self):
        self.assertFalse(isvalid('2007PECCB000'))
        self.assertFalse(isvalid('2008PECCB000'))
        self.assertFalse(isvalid('2009PECCB000'))
        self.assertFalse(isvalid('2000PECCB000'))
        self.assertFalse(isvalid('1000PECCB000'))
        self.assertFalse(isvalid('0000PECCB000'))
        self.assertFalse(isvalid('0009PECCB000'))
        self.assertFalse(isvalid('9900PECCB000'))
        self.assertFalse(isvalid('9999PECCB000'))

    def test_pec(self):
        self.assertFalse(isvalid('2021PITCB000'))
        self.assertFalse(isvalid('2021PITCB000'))
        self.assertFalse(isvalid('2021PCECB000'))
        self.assertFalse(isvalid('2021BECCB000'))
        self.assertFalse(isvalid('2021EPCCB000'))
        self.assertFalse(isvalid('2021MECCB000'))
        self.assertFalse(isvalid('2021CECCB000'))
        self.assertFalse(isvalid('2021EECCB000'))
        self.assertFalse(isvalid('2021PEXCB000'))
        
        
    # ------------------------------------------------------

    # Parse

    # ------------------------------------------------------

    def test_year(self):
        self.assertEqual(parse('2021PECCB000', 'year'), '2021')
        self.assertEqual(parse('2020PECCB000', 'year'), '2020')
        self.assertEqual(parse('2019PECCB000', 'year'), '2019')
        self.assertEqual(parse('2018PECCB000', 'year'), '2018')

    def test_dept(self):
        self.assertEqual(parse('2021PECCB000', 'department_code'), 'CB')
        self.assertEqual(parse('2021PECCS000', 'department_code'), 'CS')
        self.assertEqual(parse('2021PECEC000', 'department_code'), 'EC')
        self.assertEqual(parse('2021PECME000', 'department_code'), 'ME')
        self.assertEqual(parse('2021PECCE000', 'department_code'), 'CE')
        self.assertEqual(parse('2021PECIT000', 'department_code'), 'IT')
        self.assertEqual(parse('2021PECEE000', 'department_code'), 'EE')

    def test_shortrollno(self):
        self.assertEqual(parse('2021PECCB000', 'shortrollno'), '000')
        self.assertEqual(parse('2021PECCB001', 'shortrollno'), '001')
        self.assertEqual(parse('2021PECCB010', 'shortrollno'), '010')
        self.assertEqual(parse('2021PECCB100', 'shortrollno'), '100')
        self.assertEqual(parse('2021PECCB999', 'shortrollno'), '999')

    def test_invalid(self):
        self.assertEqual(parse('2021PECCB000', 'invalid'), None)
        self.assertEqual(parse('2021PECCB00', 'year'), None)
        self.assertEqual(parse('2021PECCB0000', 'year'), None)
        self.assertEqual(parse('2021PECB000', 'dept'), None)
        self.assertEqual(parse('2021PECCB00X', 'shortrollno'), None)
        self.assertEqual(parse('2021PECCB000', 'rollno'), None)

    # ------------------------------------------------------

    # Get Department

    # ------------------------------------------------------


    def test_get_dept(self):
        self.assertEqual(get_dept('2021PECCB000'), 'Computer Science and Business Systems')
        self.assertEqual(get_dept('2021PECCS000'), 'Computer Science and Engineering')
        self.assertEqual(get_dept('2021PECIT000'), 'Information Technology')
        self.assertEqual(get_dept('2021PECEC000'), 'Electronics and Communication Engineering')
        self.assertEqual(get_dept('2021PECEE000'), 'Electrical and Electronics Engineering')
        self.assertEqual(get_dept('2021PECME000'), 'Mechanical Engineering')
        self.assertEqual(get_dept('2021PECCE000'), 'Civil Engineering')
        self.assertEqual(get_dept('2021PECCC000'), 'Computer and Communication Engineering')
        self.assertEqual(get_dept('2021PECME000'), 'Mechanical Engineering')
        self.assertEqual(get_dept('2021PECAD000'), 'Artificial Intelligence and Data Science')

    def test_get_dept_invalid(self):
        self.assertEqual(get_dept('2021PECCB00'), None)
        self.assertEqual(get_dept('2021PECCB0000'), None)
        self.assertEqual(get_dept('2021PECB0B0'), None)
        self.assertEqual(get_dept('2021PITB000'), None)
        self.assertEqual(get_dept('2021PITCB00X'), None)
        self.assertEqual(get_dept('2021PECCB0T0'), None)

    # ------------------------------------------------------

    # Get Dept Code

    # ------------------------------------------------------
    def test_get_dept_code(self):
        self.assertEqual(get_dept_code('2021PECCB000'), 'CB')
        self.assertEqual(get_dept_code('2021PECCS000'), 'CS')
        self.assertEqual(get_dept_code('2021PECIT000'), 'IT')
        self.assertEqual(get_dept_code('2021PECEC000'), 'EC')
        self.assertEqual(get_dept_code('2021PECEE000'), 'EE')
        self.assertEqual(get_dept_code('2021PECME000'), 'ME')
        self.assertEqual(get_dept_code('2021PECCE000'), 'CE')
        self.assertEqual(get_dept_code('2021PECCC000'), 'CC')
        self.assertEqual(get_dept_code('2021PECAD000'), 'AD')

    def test_get_dept_code_invalid(self):
        self.assertEqual(get_dept_code('2021PECCB00'), None)
        self.assertEqual(get_dept_code('2021PECCB0000'), None)
        self.assertEqual(get_dept_code('2021PECB0B0'), None)
        self.assertEqual(get_dept_code('2021PITB000'), None)
        self.assertEqual(get_dept_code('2021PITCB00X'), None)
        self.assertEqual(get_dept_code('2021PECCB0T0'), None)
        



        
if __name__ == '__main__':
    unittest.main()



import regex


# Regex for rollno
regex_rollno = r'^(20[1-3]{1}[0-9]{1})(PEC)([A-Z]{2})([1-9]{1}[0-9]{3})$'

def isvalid(rollno):
    """
    Check if rollno is valid (eg. 2021PECCB001)
    Returns True if valid, False otherwise
    """

    

    # Check if rollno is valid
    if regex.match(regex_rollno, rollno.strip().upper()):
        return True
    else:
        return False


def parse(rollno, required):

    """
    Parse rollno
    Returns parsed rollno if valid, None otherwise
    Required = year, college_code, department_code, shortrollno
    """


    # Check if rollno is valid
    if isvalid(rollno):

        # Match regex
        re = regex.match(regex_rollno, rollno.strip().upper())

    
        # Return required
        if required == 'year':
            # Get year
            year = re.group(1)
            return year
        
        elif required == 'college_code':
            # Get college code
            college_code = re.group(2)
            return college_code
        
        elif required == 'department_code':
            # Get department code
            department_code = re.group(3)
            return department_code
        
        elif required == 'shortrollno':
            # Get shortrollno
            shortrollno = re.group(4)
            return shortrollno
        
        else:
            return None
    else:
        return None


def get_dept_code(rollno):
    """
    Get department code from rollno
    Returns department code if valid, None otherwise
    """

    # Check if rollno is valid
    if isvalid(rollno):

        # Get department code
        department_code = parse(rollno, 'department_code')

        return department_code
    else:
        return None


def get_dept(rollno):
    """
    Get department from rollno
    Returns department name if valid, None otherwise
    """

    # Check if rollno is valid
    if isvalid(rollno):
        departments = {
            'CB': 'Computer Science and Business Systems',
            'CS': 'Computer Science and Engineering',
            'IT': 'Information Technology',
            'EC': 'Electronics and Communication Engineering',
            'EE': 'Electrical and Electronics Engineering',
            'CE': 'Civil Engineering',
            'CC': 'Computer and Communication Engineering',
            'ME': 'Mechanical Engineering',
            'AD': 'Artificial Intelligence and Data Science'
        }

        # Get department code
        department_code = parse(rollno, 'department_code')


        # Get department name and return
        try:
            department=  departments[department_code]
        except KeyError:
            return None
        
        return department
    else:
        return None
    
    
def get_year(rollno):
    """
    Get year from rollno
    Returns year if valid, None otherwise
    """

    # Check if rollno is valid
    if isvalid(rollno):

        # Get year
        year = parse(rollno, 'year')

        return year
    else:
        return None


import pytest
from src.cgpa import *


# Subject class test cases

def test_subject_class():
    # Test if the subject class is defined
    obj = Subject(4, 10)
    assert obj.credit == 4
    assert obj.grade_point == 10

    obj = Subject(3, 9)
    assert obj.credit == 3
    assert obj.grade_point == 9

    # raise an error if the credit is not an integer
    with pytest.raises(TypeError):
        obj = Subject(3.5, 9)
    with pytest.raises(TypeError):
        obj = Subject('three', 9)
    with pytest.raises(TypeError):
        obj = Subject(True, 9)

    # raise an error if the grade_point is not an integer
    with pytest.raises(TypeError):
        obj = Subject(3, 9.5)
    with pytest.raises(TypeError):
        obj = Subject(3, 'nine')
    with pytest.raises(TypeError):
        obj = Subject(3, True)

    # raise an error if the credit and grade point is not negative
    with pytest.raises(ValueError):
        obj = Subject(-3, 9)
    with pytest.raises(ValueError):
        obj = Subject(3, -9)
    with pytest.raises(ValueError):
        obj = Subject(-3, -9)
    
# GPA class test cases

# addsubject function test cases

def test_addsubject():
    # create an object of the gpa class
    obj = Gpa()

    # Test if the addsubject function is defined
    obj.addsubject(4, 10)
    assert obj.no_of_subjects == 1
    assert obj.sub[0].credit == 4
    assert obj.sub[0].grade_point == 10

    obj.addsubject(3, 9)
    assert obj.no_of_subjects == 2
    assert obj.sub[1].credit == 3
    assert obj.sub[1].grade_point == 9


def test_invalid_credit():
    with pytest.raises(TypeError):
        obj = Subject(3.5, 9)
    with pytest.raises(TypeError):
        obj = Subject('three', 9)
    with pytest.raises(TypeError):
        obj = Subject(True, 9)


def test_invalid_grade_point():
    with pytest.raises(TypeError):
        obj = Subject(3, 9.5)
    with pytest.raises(TypeError):
        obj = Subject(3, 'nine')
    with pytest.raises(TypeError):
        obj = Subject(3, True)


def test_invalid_credit_and_grade_point():
    obj = Gpa()
    # raise an error if the credit and grade point is not negative
    with pytest.raises(ValueError):
        obj.addsubject(-3, 9)
    with pytest.raises(ValueError):
        obj.addsubject(3, -9)
    with pytest.raises(ValueError):
        obj.addsubject(-3, -9)


def test_credit_greater_than_8():
    obj = Gpa()

    # raise an error if the credit is greater than 8
    with pytest.raises(ValueError):
        obj.addsubject(9, 9)
    with pytest.raises(ValueError):
        obj.addsubject(13, 9)


def test_grade_point_greater_than_10():
    obj = Gpa()

    # raise an error if the grade point is greater than 10
    with pytest.raises(ValueError):
        obj.addsubject(3, 11)
    with pytest.raises(ValueError):
        obj.addsubject(3, 15)

# calc function test cases

def test_calc_1():
    obj = Gpa()

    # Test if the calc function is defined
    obj.addsubject(4, 10)
    obj.addsubject(3, 9)
    obj.addsubject(3, 8)
    obj.addsubject(2, 7)
    obj.addsubject(2, 8)
    obj.addsubject(1, 9)
    obj.addsubject(3, 9)
    obj.addsubject(2, 3)
    obj.calc()
    
    assert obj.no_of_subjects == 8

    assert obj.gpa == 8.722222222222221


def test_calc_2():
    obj = Gpa()

    # Test if the calc function is defined
    obj.addsubject(4, 10)
    obj.addsubject(3, 10)
    obj.addsubject(3, 10)
    obj.addsubject(2, 10)
    obj.addsubject(2, 10)
    obj.addsubject(1, 10)
    obj.addsubject(3, 10)
    obj.addsubject(2, 10)
    obj.addsubject(2, 10)
    obj.calc()

    assert obj.gpa == 10


def test_calc_3():
    obj = Gpa()

    # Test if the calc function is defined
    obj.addsubject(4, 7)
    obj.addsubject(3, 7)
    obj.addsubject(3, 7)
    obj.addsubject(2, 7)
    obj.addsubject(2, 7)
    obj.addsubject(1, 7)
    obj.addsubject(3, 7)
    obj.addsubject(2, 7)
    obj.addsubject(2, 7)
    obj.calc()

    assert obj.gpa == 7


def test_calc_4():
    obj = Gpa()

    # Test if the calc function is defined
    obj.addsubject(2, 8)
    obj.addsubject(3, 8)
    obj.addsubject(3, 7)
    obj.addsubject(4, 8)
    obj.addsubject(2, 10)
    obj.addsubject(4, 9)
    obj.addsubject(4, 9)
    obj.addsubject(3, 8)
    obj.calc()

    assert obj.gpa == 8.36


def test_calc_with_no_subjects():
    obj = Gpa()

    # Test if the calc function is defined
    with pytest.raises(ValueError):
        obj.calc()

    assert obj.gpa == 0.0

# display function test cases

def test_display(capsys):
    obj = Gpa()

    # Test if the display function is defined
    obj.addsubject(4, 10)
    obj.addsubject(3, 9)
    obj.addsubject(3, 8)
    obj.addsubject(2, 7)
    obj.addsubject(2, 7)
    obj.addsubject(1, 9)
    obj.addsubject(3, 8)
    obj.addsubject(2, 3)

    # calculate and call the display function
    obj.calc()
    obj.display()
    
    # capture the output(prints)
    captured = capsys.readouterr()

    # check if the output is correct
    assert captured.out == "No of subjects:  8\nGPA:  8.444444444444445\n"


def test_display_with_no_subjects():
    obj = Gpa()

    # Test if the display function is defined
    with pytest.raises(ValueError):
        obj.display()
    
    assert obj.gpa == 0.0


def test_display_with_no_calc():
    obj = Gpa()

    # Test if the display function is defined
    obj.addsubject(4, 10)
    obj.addsubject(3, 9)
    obj.addsubject(3, 8)
    obj.addsubject(2, 7)
    obj.addsubject(2, 7)
    obj.addsubject(1, 9)
    obj.addsubject(3, 8)
    obj.addsubject(2, 3)

    with pytest.raises(ValueError):
        obj.display()
    
    assert obj.gpa == 0.0


# gpa class overall test cases
def test_gpa_class(capsys):
    obj = Gpa()

    # add subjects
    credits = [2, 3, 3, 4, 2, 4, 4, 3]
    grade_points = [8, 8, 7, 8, 10, 9, 9, 8]

    for i in range(len(credits)):
        obj.addsubject(credits[i], grade_points[i])

    # check if the number of subjects is correct
    assert obj.no_of_subjects == 8

    # check if subjects are added to list
    assert len(obj.sub) == 8

    # check if subjects are added correctly
    for i in range(len(obj.sub)):
        assert obj.sub[i].credit == credits[i]
        assert obj.sub[i].grade_point == grade_points[i]

    # calculate
    obj.calc()

    # display
    obj.display()

    # capture the output(prints)
    captured = capsys.readouterr()

    # check if the output is correct
    assert captured.out == "No of subjects:  8\nGPA:  8.36\n"

    # check if the gpa is correct
    assert obj.gpa == 8.36


# cgpa class test cases


# addsemester function test cases
def test_addsemester():
    obj = Cgpa()

    gpa_obj = Gpa()

    # add subjects
    credits = [2, 3, 3, 4, 2, 4, 4, 3]
    grade_points = [8, 8, 7, 8, 10, 9, 9, 8]

    for i in range(len(credits)):
        gpa_obj.addsubject(credits[i], grade_points[i])


    # Test if the addsemester function is defined
    obj.addsemester(gpa_obj)

    # check if the semester is added to the list
    assert len(obj.semester) == 1

    # check if the semester is added correctly
    assert obj.semester[0].gpa == 8.36

def test_invalid_addsemester():
    obj = Cgpa()

    # Test if the addsemester function is defined
    with pytest.raises(TypeError):
        obj.addsemester(-1)

    with pytest.raises(TypeError):
        obj.addsemester(15)

    with pytest.raises(TypeError):
        obj.addsemester('hello')

    with pytest.raises(TypeError):
        obj.addsemester(3.14)

def test_addsemester_with_no_subjects():
    obj = Cgpa()

    gpa_obj = Gpa()

    # Test if the addsemester function is defined
    with pytest.raises(ValueError):
        obj.addsemester(gpa_obj)

    assert len(obj.semester) == 0

def test_addsemester_with_fail_gradepoints():
    obj = Cgpa()

    gpa_obj = Gpa()

    # add subjects
    credits = [2, 3, 3, 4, 2, 4, 4, 3]
    grade_points = [1, 2, 4, 2, 1, 2, 4, 2]

    for i in range(len(credits)):
        gpa_obj.addsubject(credits[i], grade_points[i])
    
    # Test if the addsemester function is defined
    with pytest.raises(ValueError):
        obj.addsemester(gpa_obj)

    assert len(obj.semester) == 0

def test_addsemester_multiple_semesters():
    obj = Cgpa()

    gpa_obj1 = Gpa()
    gpa_obj2 = Gpa()
    gpa_obj3 = Gpa()

    # sem 1
    credits = [2, 3, 3, 4, 2, 4, 4, 3]
    grade_points = [8, 8, 7, 8, 10, 9, 9, 8]

    for i in range(len(credits)):
        gpa_obj1.addsubject(credits[i], grade_points[i])

    # sem 2
    credits = [2, 4, 3, 4, 2, 2, 4, 2]
    grade_points = [7, 8, 7, 8, 10, 6, 9, 8]

    for i in range(len(credits)):
        gpa_obj2.addsubject(credits[i], grade_points[i])

    # sem 3
    credits = [2, 3, 4, 3, 2, 3, 4, 3]
    grade_points = [10, 8, 7, 8, 7, 9, 7, 8]

    for i in range(len(credits)):
        gpa_obj3.addsubject(credits[i], grade_points[i])

    # Test if the addsemester function is defined
    obj.addsemester(gpa_obj1)
    obj.addsemester(gpa_obj2)
    obj.addsemester(gpa_obj3)

    assert len(obj.semester) == 3

    assert obj.semester[0].gpa == 8.36
    assert obj.semester[1].gpa == 7.956521739130435
    assert obj.semester[2].gpa == 7.875



# calc function test cases
def test_cgpa_calc():
    obj = Cgpa()

    gpa_obj = Gpa()

    # add subjects
    credits = [2, 3, 3, 4, 2, 4, 4, 3]
    grade_points = [8, 8, 7, 8, 10, 9, 9, 8]

    for i in range(len(credits)):
        gpa_obj.addsubject(credits[i], grade_points[i])

    obj.addsemester(gpa_obj)

    # Test if the calc function is defined
    obj.calc()

    assert obj.cgpa == 8.36